#define CompositingMode_Normal 0
#define CompositingMode_Multiply 1
#define CompositingMode_Screen 2
#define CompositingMode_Lighten 3
#define CompositingMode_Darken 4
#define CompositingMode_Auto 5
#define CompositingMode_Dissolve 6 // Not implemented, will use Normal blending
#define CompositingMode_ColorBurn 7
#define CompositingMode_LinearBurn 8
#define CompositingMode_ColorDodge 9
#define CompositingMode_LinearDodge 10
#define CompositingMode_Overlay 11
#define CompositingMode_SoftLight 12
#define CompositingMode_HardLight 13
#define CompositingMode_VividLight 14
#define CompositingMode_LinearLight 15
#define CompositingMode_PinLight 16
#define CompositingMode_Hardmix 17
#define CompositingMode_Difference 18
#define CompositingMode_Exclusion 19
#define CompositingMode_Hue 20
#define CompositingMode_Saturation 21
#define CompositingMode_Color 22
#define CompositingMode_Luminosity 23
#define CompositingMode_Eraser 24
#define CompositingMode_Max 25


// unused macro
#define BlendRGB(base, blend, funcf) vec3(funcf(base.r, blend.r), funcf(base.g, blend.g), funcf(base.b, blend.b))
#define ColorBurnf(base, blend) ((blend > 0.0) ? max((1.0 - ((1.0 - base) / blend)), 0.0) : blend)
#define ColorDodgef(base, blend) ((blend < 1.0) ? min(base / (1.0 - blend), 1.0) : blend)
#define Overlayf(base, blend) (base < 0.5 ? (2.0 * base * blend) : (1.0 - 2.0 * (1.0 - base) * (1.0 - blend)))
#define SoftLightf(base, blend) ((blend < 0.5) ? (2.0 * base * blend + base * base * (1.0 - 2.0 * blend)) : (sqrt(base) * (2.0 * blend - 1.0) + 2.0 * base * (1.0 - blend)))
#define VividLightf(base, blend) ((blend < 0.5) ? ColorBurnf(base, (2.0 * blend)) : ColorDodgef(base, (2.0 * (blend - 0.5))))
#define Hardmixf(base, blend) ((VividLightf(base, blend) < 0.5) ? 0.0 : 1.0)
#define LinearBurnf(base, blend) max(base + blend - 1.0, 0.0)
#define LinearDodgef(base, blend) min(base + blend, 1.0)
#define LinearLightf(base, blend) (blend < 0.5 ? LinearBurnf(base, (2.0 * blend)) : LinearDodgef(base, (2.0 * (blend - 0.5))))
#define Lightenf(base, blend) max(blend, base)
#define Darkenf(base, blend) min(blend, base)
#define PinLightf(base, blend) ((blend < 0.5) ? Darkenf(base, (2.0 * blend)) : Lightenf(base, (2.0 *(blend - 0.5))))

// from https://www.adobe.com/content/dam/acom/en/devnet/acrobat/pdfs/pdf_reference_1-7.pdf
// page 523/524
lowp float luminance(lowp vec3 col) {
    return dot(col, vec3(0.3, 0.59, 0.11));
}

lowp vec3 clipColor(lowp vec3 col) {
    lowp float l = luminance(col);
    lowp float n = min(col.r, min(col.r, col.g));
    lowp float x = max(col.r, max(col.r, col.g));
    
    if(n < 0.0) {
        col = vec3(l) + (((col - vec3(l)) * l) / (l - n));
    }
    if(x > 1.0) {
        col = vec3(l) + (((col - vec3(l)) * (1.0 - l)) / (x - l));
    }
    
    return col;
}

lowp vec3 setLuminance(lowp vec3 col, lowp float l) {
    lowp float d = l - luminance(col);
    
    lowp vec3 res = col + vec3(d);
    return clipColor(res);
}

lowp float saturation(lowp vec3 col) {
    return max(col.r, max(col.g, col.b)) - min(col.r, min(col.g, col.b));
}

lowp vec3 setSaturation(lowp vec3 col, lowp float s) {
    lowp float C[3];
    C[0] = col.r,
    C[1] = col.g;
    C[2] = col.b;
    
    int iMin = col.r > col.g ? (col.g > col.b ? 2 : 1) : (col.r > col.b ? 2 : 0);
    int iMax = col.r <= col.g ? (col.g <= col.b ? 2 : 1) : (col.r <= col.b ? 2 : 0);
    int iMid = 3 - (iMax + iMin);

    if(C[iMax] > C[iMin]) {
        C[iMid] = (((C[iMid] - C[iMin]) * s) / (C[iMax] - C[iMin]));
        C[iMax] = s;
    }
    else {
        C[iMid] = 0.0;
        C[iMax] = 0.0;
    }
    
    C[iMin] = 0.0;
    
    return vec3(C[0], C[1], C[2]);
}

// blendMode is defining when compiling the shader
lowp vec4 blendLayer(lowp vec4 src, lowp vec4 dst, int blendMode, bool alphaLocked) {
    lowp float alpha = dst.a;
    
    if(alphaLocked) {
        if (alpha > 0.0)
        {
            dst = vec4((1.0 / alpha) * dst.rgb, 1.0);
        }
        else
            return vec4(0.0);
    }
    
    lowp vec4 res = vec4(0.0);

    if(blendMode == CompositingMode_Multiply || blendMode == CompositingMode_Auto) {
        res = src * (dst + vec4(1.0 - dst.a)) + vec4(1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_Screen) {
        res = src + (vec4(1) - src) * dst;
    }
    else if(blendMode == CompositingMode_Lighten) {
        res = max(src * dst.a, dst * src.a) + src * (1.0 - dst.a) + dst * (1.0 - src.a);
    }
    else if(blendMode == CompositingMode_Darken) {
        res = min(src * dst.a, dst * src.a) + src * (1.0 - dst.a) + dst * (1.0 - src.a);
    }
    else if(blendMode == CompositingMode_ColorBurn) {
        res.a = src.a * dst.a;
        res.r = src.r > 0.0 ? max(res.a - src.a * src.a * (dst.a - dst.r) / src.r, 0.0) : 0.0;
        res.g = src.g > 0.0 ? max(res.a - src.a * src.a * (dst.a - dst.g) / src.g, 0.0) : 0.0;
        res.b = src.b > 0.0 ? max(res.a - src.a * src.a * (dst.a - dst.b) / src.b, 0.0) : 0.0;
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_LinearBurn) {
        res = max(src * dst.a + (dst - dst.a) * src.a, 0.0);
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_ColorDodge) {
        res.a = src.a * dst.a;

        res.r = src.r < src.a ? min(dst.r * src.a * src.a / (src.a - src.r), res.a) : src.r * dst.a;
        res.g = src.g < src.a ? min(dst.g * src.a * src.a / (src.a - src.g), res.a) : src.g * dst.a;
        res.b = src.b < src.a ? min(dst.b * src.a * src.a / (src.a - src.b), res.a) : src.b * dst.a;
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_LinearDodge) {
        // from http://redqueengraphics.com/2018/08/11/metal-shaders-blending-basics/
        lowp vec3 rgbTest = src.rgb * dst.a + dst.rgb * src.a;
        lowp float alphaTest = src.a * dst.a;
        
        res.r = rgbTest.r >= alphaTest ? alphaTest + src.r * (1.0 - dst.a) + dst.r * (1.0 - src.a) : src.r + dst.r;
        res.g = rgbTest.g >= alphaTest ? alphaTest + src.g * (1.0 - dst.a) + dst.g * (1.0 - src.a) : src.g + dst.g;
        res.b = rgbTest.b >= alphaTest ? alphaTest + src.b * (1.0 - dst.a) + dst.b * (1.0 - src.a) : src.b + dst.b;
        res.a = src.a + dst.a - src.a * dst.a;
    }
    else if(blendMode == CompositingMode_Overlay) {
        res.a = src.a * dst.a;
        res.r = dst.r < dst.a * 0.5 ? (2.0 * src.r * dst.r) : (2.0 * (src.r * dst.a + dst.r * src.a - src.r * dst.r) - res.a);
        res.g = dst.g < dst.a * 0.5 ? (2.0 * src.g * dst.g) : (2.0 * (src.g * dst.a + dst.g * src.a - src.g * dst.g) - res.a);
        res.b = dst.b < dst.a * 0.5 ? (2.0 * src.b * dst.b) : (2.0 * (src.b * dst.a + dst.b * src.a - src.b * dst.b) - res.a);
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
        
    }
    else if(blendMode == CompositingMode_SoftLight) {
        res.a = src.a * dst.a;
        res.r = src.r < src.a * 0.5 ? (dst.a > 0.0 ? 2.0 * dst.r * src.r + dst.r * dst.r / dst.a * (src.a - 2.0 * src.r) : 0.0) : sqrt(dst.r * dst.a) * (2.0 * src.r - src.a) + 2.0 * dst.r * (src.a - src.r);
        res.g = src.g < src.a * 0.5 ? (dst.a > 0.0 ? 2.0 * dst.g * src.g + dst.g * dst.g / dst.a * (src.a - 2.0 * src.g) : 0.0) : sqrt(dst.g * dst.a) * (2.0 * src.g - src.a) + 2.0 * dst.g * (src.a - src.g);
        res.b = src.b < src.a * 0.5 ? (dst.a > 0.0 ? 2.0 * dst.b * src.b + dst.b * dst.b / dst.a * (src.a - 2.0 * src.b) : 0.0) : sqrt(dst.b * dst.a) * (2.0 * src.b - src.a) + 2.0 * dst.b * (src.a - src.b);
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_HardLight) {
        res.a = src.a * dst.a;
        res.r = src.r < src.a * 0.5 ? (2.0 * src.r * dst.r) : (2.0 * (src.r * dst.a + dst.r * src.a - src.r * dst.r) - res.a);
        res.g = src.g < src.a * 0.5 ? (2.0 * src.g * dst.g) : (2.0 * (src.g * dst.a + dst.g * src.a - src.g * dst.g) - res.a);
        res.b = src.b < src.a * 0.5 ? (2.0 * src.b * dst.b) : (2.0 * (src.b * dst.a + dst.b * src.a - src.b * dst.b) - res.a);
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_VividLight) {
        res.a = src.a * dst.a;
        res.r = src.r < src.a * 0.5 ? (src.r > 0.0 ? max(res.a - src.a * src.a * (dst.a - dst.r) / (2.0 * src.r), 0.0) : 0.0) : (src.r < src.a ? min(dst.r * src.a * src.a / (2.0 * (src.a - src.r)), res.a) : (2.0 * src.r - src.a) * dst.a);
        res.g = src.g < src.a * 0.5 ? (src.g > 0.0 ? max(res.a - src.a * src.a * (dst.a - dst.g) / (2.0 * src.g), 0.0) : 0.0) : (src.g < src.a ? min(dst.g * src.a * src.a / (2.0 * (src.a - src.g)), res.a) : (2.0 * src.g - src.a) * dst.a);
        res.b = src.b < src.a * 0.5 ? (src.b > 0.0 ? max(res.a - src.a * src.a * (dst.a - dst.b) / (2.0 * src.b), 0.0) : 0.0) : (src.b < src.a ? min(dst.b * src.a * src.a / (2.0 * (src.a - src.b)), res.a) : (2.0 * src.b - src.a) * dst.a);
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_LinearLight) {
        res.a = src.a * dst.a;
        res.r = src.r < src.a * 0.5 ? max(dst.r * src.a + 2.0 * src.r * dst.a - res.a, 0.0) : min(dst.r * src.a + 2.0 * src.r * dst.a - res.a, 1.0);
        res.g = src.g < src.a * 0.5 ? max(dst.g * src.a + 2.0 * src.g * dst.a - res.a, 0.0) : min(dst.g * src.a + 2.0 * src.g * dst.a - res.a, 1.0);
        res.b = src.b < src.a * 0.5 ? max(dst.b * src.a + 2.0 * src.b * dst.a - res.a, 0.0) : min(dst.b * src.a + 2.0 * src.b * dst.a - res.a, 1.0);
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_PinLight) {
        res.a = src.a * dst.a;
        res.r = src.r < src.a * 0.5 ? min(dst.r * src.a, 2.0 * src.r * dst.a) : max(dst.r * src.a, 2.0 * src.r * dst.a - res.a);
        res.g = src.g < src.a * 0.5 ? min(dst.g * src.a, 2.0 * src.g * dst.a) : max(dst.g * src.a, 2.0 * src.g * dst.a - res.a);
        res.b = src.b < src.a * 0.5 ? min(dst.b * src.a, 2.0 * src.b * dst.a) : max(dst.b * src.a, 2.0 * src.b * dst.a - res.a);
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_Hardmix) {
        res.a = src.a * dst.a;
        
        // vivid light
        res.r = src.r < src.a * 0.5 ? (src.r > 0.0 ? max(res.a - src.a * src.a * (dst.a - dst.r) / (2.0 * src.r), 0.0) : 0.0) : (src.r < src.a ? min(dst.r * src.a * src.a / (2.0 * (src.a - src.r)), res.a) : (2.0 * src.r - src.a) * dst.a);
        res.g = src.g < src.a * 0.5 ? (src.g > 0.0 ? max(res.a - src.a * src.a * (dst.a - dst.g) / (2.0 * src.g), 0.0) : 0.0) : (src.g < src.a ? min(dst.g * src.a * src.a / (2.0 * (src.a - src.g)), res.a) : (2.0 * src.g - src.a) * dst.a);
        res.b = src.b < src.a * 0.5 ? (src.b > 0.0 ? max(res.a - src.a * src.a * (dst.a - dst.b) / (2.0 * src.b), 0.0) : 0.0) : (src.b < src.a ? min(dst.b * src.a * src.a / (2.0 * (src.a - src.b)), res.a) : (2.0 * src.b - src.a) * dst.a);
        
        // hardmix
        res.rgb = step(0.5 * res.a, res.rgb) * res.a;
        
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_Difference) {
        res.a = src.a * dst.a;
        res.rgb = abs(dst.rgb * src.a - src.rgb * dst.a);
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_Exclusion) {
        res.a = src.a * dst.a;
        res.rgb = dst.rgb * src.a + src.rgb * dst.a - 2.0 * dst.rgb * src.rgb;
        res += (1.0 - dst.a) * src + (1.0 - src.a) * dst;
    }
    else if(blendMode == CompositingMode_Hue) {
        if(dst.a > 0.0) {
            if(src.a > 0.0) {
                src.rgb = src.rgb / src.a;
                dst.rgb = dst.rgb / dst.a;

                res.rgb = setLuminance(setSaturation(src.rgb, saturation(dst.rgb)), luminance(dst.rgb)) * src.a * dst.a;
                res.a = src.a * dst.a;
                
                res += (1.0 - dst.a) * src * src.a;
                res += (1.0 - src.a) * dst * dst.a;
            }
            else {
                res = dst;
            }
        }
        else {
            res = src;
        }
    }
    else if(blendMode == CompositingMode_Saturation) {
        if(dst.a > 0.0) {
            if(src.a > 0.0) {
                src.rgb = src.rgb / src.a;
                dst.rgb = dst.rgb / dst.a;

                res.rgb = setLuminance(setSaturation(dst.rgb, saturation(src.rgb)), luminance(dst.rgb)) * src.a * dst.a;
                res.a = src.a * dst.a;
                
                res += (1.0 - dst.a) * src * src.a;
                res += (1.0 - src.a) * dst * dst.a;
            }
            else {
                res = dst;
            }
        }
        else {
            res = src;
        }
    }
    else if(blendMode == CompositingMode_Color) {
        if(dst.a > 0.0) {
            if(src.a > 0.0) {
                src.rgb = src.rgb / src.a;
                dst.rgb = dst.rgb / dst.a;

                res.rgb = setLuminance(src.rgb, luminance(dst.rgb)) * src.a * dst.a;
                res.a = src.a * dst.a;
                
                res += (1.0 - dst.a) * src * src.a;
                res += (1.0 - src.a) * dst * dst.a;
            }
            else {
                res = dst;
            }
        }
        else {
            res = src;
        }
    }
    else if(blendMode == CompositingMode_Luminosity) {
        if(dst.a > 0.0) {
            if(src.a > 0.0) {
                src.rgb = src.rgb / src.a;
                dst.rgb = dst.rgb / dst.a;

                res.rgb = setLuminance(dst.rgb, luminance(src.rgb)) * src.a * dst.a;
                res.a = src.a * dst.a;
                    
                res += (1.0 - dst.a) * src * src.a;
                res += (1.0 - src.a) * dst * dst.a;
            }
            else {
                res = dst;
            }
        }
        else {
            res = src;
        }
    }
    else if(blendMode == CompositingMode_Eraser) {
        res = dst * (1.0 - src.a);
    }
    else if(blendMode == CompositingMode_Max) {
        res = src.a > dst.a ? src : dst;
    }
    else { // use CompositingMode_Normal
        res = src + (1.0 - src.a) * dst;
    }
    
    if(alphaLocked) {
        res *= alpha;
    }
    
    return res;
}
